#!/usr/bin/env python
'''
Command-line utility for calculating chemical features.
'''

import argparse
import contextlib
import logging
import pathlib
import sys

import yaml
import pandas as pd

from chemfeat.database import FeatureDatabase
from chemfeat.features.manager import FeatureManager, import_calculators


LOGGER = logging.getLogger(__name__)


@contextlib.contextmanager
def open_file_or_stdout(path=None, log_msg=None):
    '''
    Context manager for retrieving a handle to an open file or STDOUT.

    Args:
        path:
            An optional file path. It will be created and opened if write mode
            if given.

        log_msg:
            An optional log message to log as INFO when a path is given. It
            should include a single placeholder for the file path.

    Returns:
        The open file handle or the STDOUT object.
    '''
    if path is None:
        yield sys.stdout
    else:
        path = pathlib.Path(path).resolve()
        if log_msg:
            LOGGER.info(log_msg, path)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open('w', encoding='utf-8') as handle:
            yield handle


def calculate_features(pargs):
    '''
    Calculate chemical features for the requested InChis.
    '''
    LOGGER.info('Calculating chemical features.')
    inchi_column = pargs.inchi
    in_path = pargs.input.resolve()
    LOGGER.info('Loading InChis from %s', in_path)
    inchis = pd.read_csv(in_path)[inchi_column]

    if inchis.empty:
        LOGGER.warning('No InChis found in %s', in_path)
        return 1

    feat_specs_path = pargs.feat_spec
    with feat_specs_path.open('r', encoding='utf-8') as handle:
        feat_specs = yaml.safe_load(handle)

    feat_db_path = pargs.database.resolve()
    feat_db_path.parent.mkdir(parents=True, exist_ok=True)
    feat_db = FeatureDatabase(feat_db_path)

    out_path = pargs.output.resolve()
    out_path.parent.mkdir(parents=True, exist_ok=True)

    feat_man = FeatureManager(feat_db, feat_specs)
    if not feat_man.feature_calculators:
        LOGGER.error('No valid features sets were configured: %s', feat_specs_path)
        return 1
    feat_man.calculate_features(inchis, output_path=out_path)
    return 0


def _descibe_lines():
    '''
    '''
    yield '# Feature Sets'
    yield ''
    for _name, calc in sorted(import_calculators().items()):
        yield from calc.get_description(level=2)


def describe_feature_sets(pargs):
    '''
    Generate descriptions of each feature set.
    '''
    LOGGER.info('Generating feature set descriptions.')
    lines = _descibe_lines()
    with open_file_or_stdout(
        path=pargs.output,
        log_msg='Saving descriptions to %s'
    ) as handle:
        for line in lines:
            handle.write(f'{line}\n')
    return 0


def configure_feature_sets(pargs):
    '''
    Generate a commented feature-set configuration file.
    '''
    with open_file_or_stdout(
        path=pargs.output,
        log_msg='Saving configuration file to %s'
    ) as handle:
        for name, calc in sorted(import_calculators().items()):
            obj = calc()
            handle.write(
                f'# {calc.get_short_description()}\n'
                f'# - name: {name}\n'
            )
            params = obj.parameters
            if params:
                for line in yaml.dump(params).strip().split('\n'):
                    handle.write(f'#   {line}\n')
            handle.write('\n')
    return 0


def parse_args(args=None):
    '''
    Parse command-line arguments.

    Args:
        args:
            Command-line arguments. If None, sys.argv is used.

    Returns:
        The command-line arguments parsed with argparse.

    TODO:
        Add option to configure feature calculator import paths.
    '''
    parser = argparse.ArgumentParser(
        description='Calculate chemical features.'
    )
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help='Enable debugging messages.'
    )
    subparsers = parser.add_subparsers(
        title='Subcommands',
        description='Available subcommands',
        dest='command'
    )

    # ------------------------------ calculate ------------------------------- #
    calc_parser = subparsers.add_parser(
        'calculate',
        aliases=['calc'],
        help='Calculate chemical features from InChis.'
    )
    calc_parser.add_argument(
        'feat_spec',
        metavar='<YAML file>',
        type=pathlib.Path,
        help='A YAML file specifying the features to calculate.'
    )
    calc_parser.add_argument(
        'input',
        metavar='<CSV file>',
        type=pathlib.Path,
        help=('''
            A CSV file containing a column with the InChis of the molecules for
            which the features should be calulated.'
        ''')
    )
    calc_parser.add_argument(
        'output',
        metavar='<CSV file>',
        type=pathlib.Path,
        help=('''
              The path to which the calculated features should be saved as a CSV
              file.
        ''')
    )
    calc_parser.add_argument(
        '-i', '--inchi',
        metavar='<InChi column name>',
        default='InChi',
        help=('''
            The name of the column containing InChi values in the input file. It
            will also be used in the output file. Default: %(default)s
        ''')
    )
    calc_parser.add_argument(
        '-d', '--database',
        metavar='<SQLite database file>',
        default='features.sqlite',
        type=pathlib.Path,
        help=('''
            The path to an SQLite database in which to cache calculated features
            for re-use. Default: %(default)s
        ''')
    )
    calc_parser.set_defaults(func=calculate_features)

    # ------------------------------ configure ------------------------------- #
    conf_parser = subparsers.add_parser(
        'configure',
        aliases=['conf'],
        help='Generate a template configuration file.'
    )
    conf_parser.add_argument(
        '-o', '--output',
        metavar='<YAML file>',
        type=pathlib.Path,
        help='Save the generated template to <YAML file>.'
    )
    conf_parser.set_defaults(func=configure_feature_sets)

    # ------------------------------- describe ------------------------------- #
    desc_parser = subparsers.add_parser(
        'describe',
        aliases=['desc'],
        help='Describe the available chemical features.'
    )
    desc_parser.add_argument(
        '-o', '--output',
        metavar='<Markdown file>',
        type=pathlib.Path,
        help='Save the output to <Markdown file>.'
    )
    desc_parser.set_defaults(func=describe_feature_sets)


    return parser.parse_args(args=args)


def main(args=None):
    '''
    Parse command-line arguments and perform requested functions.

    Args:
        Passed through to parse_args().
    '''
    pargs = parse_args(args=args)
    logging.basicConfig(
        style='{',
        format='[{asctime}] {levelname} {message}',
        datefmt='%Y-%m-%d %H:%M:%S',
        level=(logging.DEBUG if pargs.debug else logging.INFO)
    )
    return pargs.func(pargs)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except (KeyboardInterrupt, BrokenPipeError):
        pass



#!/usr/bin/env python3

'''
Morgan fingerprint feature calculator.
'''


from rdkit.Chem import AllChem


from chemfeat.features.calculator import FeatureCalculator


class ECFPFeatureCalculator(FeatureCalculator):
    '''
    ECFP feature calculator.

    Extended-Connectivity fingerprints (ECFP), a.k.a. Morgan fingerprints,
    calculated with [RDKit cheminformatics
    library](http://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html?highlight=getmorganfingerprintasbitvect#rdkit.Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect)

    > Rogers, David, and Mathew Hahn. “Extended-Connectivity Fingerprints.”
    > Journal of Chemical Information and Modeling 50, no. 5 (May 24, 2010):
    > 742–54. https://doi.org/10.1021/ci100050t.

    Each feature is a single bit of the feature vector.
    '''
    FEATURE_SET_NAME = 'ecfp'

    def __init__(self, size: int = 2048):
        '''
        Args:
            size:
                The fingerprint size. It should be 1024, 2048 or 4096.
        '''
        size = int(size)
        valid_sizes = (1024, 2048, 4096)
        if size not in valid_sizes:
            raise ValueError(
                f'Invalid size for fingerprint: {size}. '
                f'Valid sizes: {valid_sizes}'
            )
        self.size = 2048

    @property
    def parameters(self):
        return {'size': self.size}

    def is_numeric(self, _name):
        return False

    def add_features(self, features, molecule):
        fingerprint = AllChem.GetMorganFingerprintAsBitVect(molecule, 2, nBits=self.size)
        return self.add_fingerprint_features(features, fingerprint)


ECFPFeatureCalculator.register()



#!/usr/bin/env python3
'''
Functions for dynamically importing modules.
'''

import logging
import pathlib
import pkgutil


LOGGER = logging.getLogger(__name__)


def import_modules(paths, path_log_msg=None):
    '''
    Import modules by paths.

    Args:
        paths:
            An iterable of file and directory paths. File paths will import
            single modules while directory paths will recursively import all
            modules contained in the directory.

        path_log_msg:
            An optional logging format string for logging messages about the
            given paths. If given, it should contain a single "%s" placeholder
            for the path argument.

    Returns:
        The set of loaded module names.
    '''
    paths = set(pathlib.Path(path).resolve() for path in paths)
    if not paths:
        LOGGER.warning('No paths passed to import_modules().')

    if path_log_msg:
        for path in sorted(paths):
            LOGGER.debug(path_log_msg, path)

    file_paths = set()
    dir_paths = set()
    for path in paths:
        if path.is_dir():
            dir_paths.add(path)
        else:
            file_paths.add(path)

    all_dir_paths = sorted(dir_paths | set(path.parent for path in file_paths))

    loaded = set()
    for loader, name, _is_pkg in pkgutil.walk_packages(path=all_dir_paths):
        path = pathlib.Path(loader.path) / f'{name}.py'
        if path in file_paths or any(path.is_relative_to(p) for p in dir_paths):
            LOGGER.debug('Loading module %s from %s', name, path)
            loader.find_module(name).load_module(name)
            loaded.add(name)

    return loaded



#!/usr/bin/env python3

'''
Base clase for feature calculators.


TODO:
    * Add methods for describing the features for automatically generating
      documentation.
'''

import inspect
import logging
from collections import OrderedDict

from docstring_parser import parse as parse_docstring


LOGGER = logging.getLogger(__name__)

# Dictionary mapping feature set names to their corresponding feature
# calculators. Each subclass will add itself.
FEATURE_CALCULATORS = {}


PREFIX_SEPARATOR = '__'


def split_prefix(name):
    '''
    Split a fully qualified feature name into a feature set name and an an unqualified feature name.
    '''
    return name.split(PREFIX_SEPARATOR, 1)


def _describe_func_args(func, header=None):
    '''
    Describe the arguments of a function.

    Args:
        func:
            The input function.

        header:
            An optional section header with preceding number signs to emit
            before the parameter list.

    Returns:
        A generator over markdown strings.
    '''
    docstring = parse_docstring(func.__doc__)
    param_descs = {param.arg_name: param.description for param in docstring.params}
    argspec = inspect.getfullargspec(func)
    n_args = len(argspec.args)

    required_args = n_args - len(argspec.defaults) \
        if argspec.defaults is not None \
        else n_args

    defaults = argspec.kwonlydefaults
    if defaults is None:
        defaults = {}
    defaults.update({
        arg: argspec.defaults[i - required_args]
        for i, arg in enumerate(argspec.args)
        if i >= required_args
    })

    all_args = [*argspec.args, *argspec.kwonlyargs]

    is_first = True
    for arg in all_args:
        if arg == 'self':
            continue
        annotations = argspec.annotations[arg]
        desc = f'* {arg}'
        specs = OrderedDict()
        if annotations:
            specs['type'] = annotations.__name__
        try:
            specs['default'] = defaults[arg]
        except KeyError:
            pass
        if specs:
            specs = '; '.join(f'{key}: {value}' for key, value in specs.items())
            desc += f' ({specs})'
        desc += f': {param_descs.get(arg, "Description unavailable.")}'

        if is_first and header:
            yield header
            yield ''
        yield desc
        is_first = False

    if not is_first:
        yield '\n\n'


class FeatureCalculator():
    '''
    Base class for feature calculators.
    '''
    FEATURE_SET_NAME = NotImplemented

    @classmethod
    def register(cls):
        '''
        Register this FeatureCalculator subclass among the available feature
        calculators.
        '''
        prev_cls = FEATURE_CALCULATORS.get(cls.FEATURE_SET_NAME)
        if prev_cls is None:
            LOGGER.debug(
                'Registering feature calculator %s: %s',
                cls.FEATURE_SET_NAME,
                cls
            )
        else:
            LOGGER.warning(
                'Reregistering feature calculator %s: %s -> %s',
                cls.FEATURE_SET_NAME,
                prev_cls,
                cls
            )
        FEATURE_CALCULATORS[cls.FEATURE_SET_NAME] = cls

    @classmethod
    def get_description(cls, level=1):
        '''
        A description of the feature set with relevant references and a
        description of the recognized parameters.

        Args:
            level:
                Markdown header level for the returned description.

        Returns:
            The description as a markdown string.
        '''
        level = max(level, 1)
        header_signs = '#' * level

        cls_docstring = parse_docstring(cls.__doc__)
        cls_desc = [
            cls_docstring.short_description,
            cls_docstring.long_description
        ]
        cls_desc = [desc for desc in cls_desc if desc]
        if cls_desc:
            cls_desc = '\n\n'.join(cls_desc)
        if not cls_desc:
            LOGGER.warning('No class description for %s', cls.__name__)
            cls_desc = 'No description available.'

        yield f'''{header_signs} {cls.FEATURE_SET_NAME}

{cls_desc}

'''
        yield from _describe_func_args(cls.__init__, header=f'{header_signs}# Parameters')

    @classmethod
    def get_short_description(cls):
        '''
        The short description from the class' docstring.

        Returns:
            The short description as a string.
        '''
        return parse_docstring(cls.__doc__).short_description

    @property
    def identifier(self):
        '''
        Get a unique ID for this feature set with the given parameters.
        '''
        params = self.parameters
        if not params:
            return self.FEATURE_SET_NAME
        param_str = '-'.join(
            f'{name}:{value}'
            for name, value in sorted(params.items())
            if value is not None
        )
        return f'{self.FEATURE_SET_NAME}-{param_str}'

    @property
    def parameters(self):
        '''
        Return a dictionary representing the parametes required to re-initialize
        the feature calculator.
        '''
        return {}

    def add_prefix(self, name):
        '''
        Add the feature set prefix to the feature name.

        Args:
            name:
                The name of a feature in the feature set handled by a subclass
                of this class.

        Returns:
            The feature name with the feature set previx.
        '''
        return f'{self.identifier}{PREFIX_SEPARATOR}{name}'

    def remove_prefix(self, name):
        '''
        Remove the feature set prefix from the feature name.

        Args:
            name:
                The name, with or without the feature set prefix.

        Returns:
            The feature name without the feature set prefix.
        '''
        return split_prefix(name)[1]

    def is_numeric(self, name):
        '''
        Return True if the given feature name is numeric and not categoric.
        '''

    @property
    def is_3D(self):
        '''
        True if the features depend on 3D molecular information.
        '''
        return False

    def add_features(self, features, molecule):
        '''
        Args:
            features:
                A dict to which to add the calculated features. The keys should
                be prefixed feature names as returned by add_prefix() and the
                values should be there value.

            molecule:
                The RDKit molecule object for which to calculate features.

        Returns:
            The updated features dict.
        '''
        raise NotImplementedError(f'{self.__class__.__name__} does not implemented add_features()')

    def add_fingerprint_features(self, features, fingerprint, name=''):
        '''
        Convenience method for adding fingerprint vectors as separate features.

        Args:
            features:
                Same as add_features().

            fingerprint:
                An iterable of values to add as separate features. These will be
                named sequentially.

            name:
                An optional name to insert between the prefix and the index.

        Returns:
            The updated features dict.
        '''
        size = len(fingerprint)
        n_digits = len(str(size - 1))
        names = list(self.add_prefix(f'{name}{i:0{n_digits:d}d}') for i in range(size))
        features.update(zip(names, fingerprint))
        return features



#!/usr/bin/env python3

'''
PaDEL feature calculators.

Given the number of PaDEL descriptors and fingerprints, this module generates
the corresponding classes programmatically.
'''

import logging
import sys
import textwrap
from typing import Optional

from PaDEL_pywrapper import PaDEL
from PaDEL_pywrapper.descriptor import descriptors, _fingerprints

from chemfeat.features.calculator import FeatureCalculator
from chemfeat.markdown import escape as escape_markdown


LOGGER = logging.getLogger(__name__)
THIS_MODULE = sys.modules[__name__]


def _get_class_name(desc):
    '''
    Get a class name for the given PaDEL descriptor or fingerprint.
    '''
    try:
        return f'padel_{desc.short_name}'
    except AttributeError:
        return f'padel_{desc.name}'


class _PaDELCommonFeatureCalculator(FeatureCalculator):  # pylint: disable=abstract-method
    '''
    Abstract base class for other PaDEL feature calculators.
    '''
    _IS_3D = False

    @property
    def is_3D(self):
        return self._IS_3D


class PaDELDescFeatureCalculator(_PaDELCommonFeatureCalculator):
    '''
    Base class for PaDEL feature calculator subclasses.
    '''
    _PADEL_DESC = NotImplemented

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.padel = PaDEL([self._PADEL_DESC])

    def is_numeric(self, _name):
        return True

    def add_features(self, features, molecule):
        values = self.padel.calculate([molecule], show_banner=False)
        for key, value in values.iloc[0].to_dict():
            features[self.add_prefix(key)] = value
        return features


class PaDELFPFeatureCalculator(_PaDELCommonFeatureCalculator):
    '''
    Base class for PaDEL fingerprint calculator subclasses.
    '''
    _PADEL_DESC = NotImplemented

    def __init__(
        self,
        *args,
        size: Optional[int] = None,
        search_depth: Optional[int] = None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._padel_kwargs = {}
        if size is not None:
            self._padel_kwargs['size'] = int(size)
        if search_depth is not None:
            self._padel_kwargs['searchDepth'] = int(search_depth)
        self.padel = PaDEL([self._PADEL_DESC], **self._padel_kwargs)

    @property
    def parameters(self):
        return self._padel_kwargs.copy()

    def is_numeric(self, _name):
        return False

    def add_features(self, features, molecule):
        fingerprint = self.padel.calculate([molecule], show_banner=False).iloc[0]
        return self.add_fingerprint_features(features, fingerprint)


def _add_docstring_and_class(cls, name, docstring, features):
    citation_blurb = '''
All values are calculated with the [PaDEL_pywrapper Python
package](https://github.com/OlivierBeq/PaDEL_pywrapper) based on
[PaDEL-descriptor](https://doi.org/10.1002/jcc.21707).

> Yap, Chun Wei. “PaDEL-Descriptor: An Open Source Software to Calculate
> Molecular Descriptors and Fingerprints.” Journal of Computational Chemistry
> 32, no. 7 (May 2011): 1466–74. https://doi.org/10.1002/jcc.21707.
'''.strip()

    docstring = '\n\n'.join(
        '\n'.join(textwrap.wrap(block, width=80))
        for block in docstring.split('\n\n')
    )
    docstring += f'\n\n{features}\n\n{citation_blurb}'
    setattr(cls, '__doc__', docstring)

    setattr(THIS_MODULE, name, cls)
    cls.register()


def declare_classes():
    '''
    Declare the descriptor and fingerprint classes.
    '''
    for desc in descriptors:
        name = _get_class_name(desc)
        dcls = type(
            name,
            (PaDELDescFeatureCalculator, ),
            {
                '_PADEL_DESC': desc,
                '_IS_3D': desc.is_3D,
                'FEATURE_SET_NAME': name
            }
        )
        qualifier = '3D ' if desc.is_3D else ''
        features = '\n'.join(
            f'* {escape_markdown(name)}: {escape_markdown(desc)}'
            for name, desc in zip(desc.description.name, desc.description.description)
        )
        docstring = f'''{desc.name} {qualifier}PaDEL descriptor

The following features are calculated:'''
        _add_docstring_and_class(dcls, name, docstring, features)

    for pfp in _fingerprints:
        name = _get_class_name(pfp)
        fpcls = type(
            name,
            (PaDELFPFeatureCalculator, ),
            {
                '_PADEL_DESC': pfp,
                '_IS_3D': pfp.is_3D,
                'FEATURE_SET_NAME': name
            }
        )
        qualifier = '3D ' if pfp.is_3D else ''
        n_bits = pfp.nBits if pfp.nBits else 'variable'
        features = f'''* Number of bits: {n_bits}
* Bit prefix: {pfp.bit_prefix}
'''
        docstring = f'''{qualifier}PaDEL {pfp.short_name}fingerprint

{pfp.name} - {pfp.description.iloc[0]}
'''
        _add_docstring_and_class(fpcls, name, docstring, features)


declare_classes()



#!/usr/bin/env python3
'''
Database functions.
'''

import json
import logging
import pathlib
import sqlite3

import pandas as pd

from chemfeat.csv_wrapper import CSVWrapper
from chemfeat.features.calculator import PREFIX_SEPARATOR


LOGGER = logging.getLogger(__name__)


def dict_factory(cursor, row):
    '''
    Row factor for SQLite to convert rows to dicts.
    '''
    fields = [column[0] for column in cursor.description]
    return dict(zip(fields, row))


class FeatureDatabase():
    '''
    Database for saving features.
    '''
    INCHI_COLUMN_NAME = 'InChi'
    INCHI_COLUMN_INFO = (INCHI_COLUMN_NAME, 'TEXT', False, None, True)
    FEATURE_NAMES_TABLE_NAME = 'feature_names'
    FEATURE_NAMES_TABLE_INFO = [
        ('table', 'TEXT', False, None, True),
        ('feature_names', 'TEXT', False, None, False),
    ]
    FEATURES_COLUMN_NAME = 'features'
    NO_SUCH_TABLE_ERROR = 'no such table:'

    def __init__(self, path):
        self.path = pathlib.Path(path).resolve()
        self.path.parent.mkdir(parents=True, exist_ok=True)
        LOGGER.info('Using database at %s', self.path)
        self.conn = sqlite3.Connection(str(self.path))

    def _get_table_info(self, table):
        '''
        Get the table information using the table_info pragma.

        Args:
            table:
                The name of the table.

        Returns:
            An iterator over the rows ("name", "type", "notnull", "dflt_value", "pk").

        https://www.sqlite.org/pragma.html#pragma_table_info
        '''
        with self.conn as conn:
            for row in conn.execute(f'PRAGMA table_info("{table}")'):
                # Skip the index.
                yield (row[1], row[2], bool(row[3]), row[4], bool(row[5]))

    @staticmethod
    def _col_info_to_col_def(col_info):
        '''
        Create a column definition from a row of table information.

        Args:
            col_info:
                The row of column information.

        Returns:
            A string with the column definition for a CREATE TABLE statement.
        '''
        name = col_info[0]
        typ = col_info[1].upper()
        notnull = ' NOT NULL' if col_info[2] else ''
        default = f' DEFAULT {col_info[3]}' if col_info[3] is not None else ''
        prim_key = ' PRIMARY KEY' if col_info[4] else ''
        return f'"{name}" {typ}{prim_key}{notnull}{default}'

    @classmethod
    def _table_info_to_col_def(cls, table_info):
        '''
        Convert table information to column definitions.

        Args:
            table_info:
                The rows of column information.

        Returns:
            A string with the column definitions for a CREATE TABLE statement.
        '''
        return ', '.join(cls._col_info_to_col_def(row) for row in table_info)

    def create_table_from_info(self, name, table_info):
        '''
        Create a table using the given table info. If the table already exists
        with the given info then no action is performed. If a table of the same
        name already exists with mismatched info then it is dropped and
        recreated.

        Args:
            name:
                The table name.

            table_info:
                The rows of table info.
        '''
        existing_info = list(self._get_table_info(name))
        if existing_info:
            if existing_info == table_info:
                return
            LOGGER.warning('Dropping table %s due to mismatched columns', name)
            with self.conn as conn:
                conn.execute(f'DROP TABLE "{name}"')

        LOGGER.info('Creating table %s', name)
        col_def = self._table_info_to_col_def(table_info)
        with self.conn as conn:
            conn.execute(f'CREATE TABLE IF NOT EXISTS "{name}" ({col_def})')

    @staticmethod
    def _group_features(features):
        '''
        Group sets of features by name.

        Args:
            features:
                A dictionary mapping features to values. All feature names must
                include a feature set prefix that ends with the
                chemfeat.features.calculator.PREFIX_SEPARATOR value. The prefix
                will be used as the feature set names.

        Returns:
            A dict mapping feature set names to the sorted names of the features in the
            set (e.g. {"foo": ["foo__feat1", "foo__feat2"]}).
        '''
        groups = {}
        for feat_name in features.keys():
            group_name = feat_name.split(PREFIX_SEPARATOR, 1)[0]
            groups.setdefault(group_name, set()).add(feat_name)
        return {name: sorted(feats) for (name, feats) in groups.items()}

    def create_feature_table(self, table_name):
        '''
        Create a table for holding a feature set. The features will be held in a
        blob due to default limits on the number of columns in an SQLite
        database.

        Args:
            table_name:
                The table name.
        '''
        table_info = [
            self.INCHI_COLUMN_INFO,
            (self.FEATURES_COLUMN_NAME, 'BLOB', False, None, False)
        ]
        self.create_table_from_info(table_name, table_info)

    def create_required_tables(self):
        '''
        Create the required tables.
        '''
        self.create_table_from_info(self.FEATURE_NAMES_TABLE_NAME, self.FEATURE_NAMES_TABLE_INFO)

    def insert_feature_names(self, set_name, feature_names):
        '''
        Insert the feature names into the database.

        Args:
            set_name:
                The feature set name.

            feature_names:
                The list of feature names.
        '''
        with self.conn as conn:
            conn.execute(
                f'REPLACE INTO {self.FEATURE_NAMES_TABLE_NAME} VALUES(?, ?)',
                [set_name, '\n'.join(feature_names)]
            )

    def get_feature_names(self, set_name):
        '''
        Get the list of feature names for the given feature set.
        '''
        try:
            with self.conn as conn:
                for row in conn.execute(
                    f'SELECT {self.FEATURE_NAMES_TABLE_INFO[1][0]} '
                    f'FROM {self.FEATURE_NAMES_TABLE_NAME} '
                    f'WHERE {self.FEATURE_NAMES_TABLE_INFO[0][0]}=?',
                    [set_name]
                ):
                    return row[0].splitlines()
        except sqlite3.OperationalError as err:
            if not err.args[0].startswith(self.NO_SUCH_TABLE_ERROR):
                LOGGER.error('%s', err)
        return None

    def get_all_feature_names(self):
        '''
        Get a dict mapping feature sets to feature names.
        '''
        feature_names = {}
        try:
            with self.conn as conn:
                for row in conn.execute(f'SELECT * FROM {self.FEATURE_NAMES_TABLE_NAME}'):
                    feature_names[row[0]] = row[1].splitlines()
        except sqlite3.OperationalError as err:
            if not err.args[0].startswith(self.NO_SUCH_TABLE_ERROR):
                LOGGER.error('%s', err)
        return feature_names

    def _parse_inserts(self, items):
        '''
        Internal generator for inserting features into table. This is used to
        accumulate rows for chunked insertion with executemany().

        Args:
            items:
                Same as insert_features.

        Returns:
            A generator over 3-tuples of database names, InChi strings and
            feature objects as JSON strings.
        '''
        is_first = True
        for inchi, features in items:
            if not features:
                LOGGER.warning('No features for %s', inchi)
                continue
            grouped_features = self._group_features(features)
            features[self.INCHI_COLUMN_NAME] = inchi

            # Ensure that the expected tables exist.
            if is_first:
                self.create_required_tables()
                all_existing_feature_names = self.get_all_feature_names()

                for name, feat_names in grouped_features.items():
                    existing_feature_names = all_existing_feature_names.get(name)
                    if existing_feature_names:
                        if existing_feature_names != feat_names:
                            LOGGER.warning(
                                "Dropping table %s due to mismatched columns",
                                name
                            )
                            with self.conn as conn:
                                conn.execute(f'DROP TABLE IF EXISTS "{name}"')
                            self.insert_feature_names(name, feat_names)
                    else:
                        self.insert_feature_names(name, feat_names)
                    self.create_feature_table(name)
                is_first = False

            for name, feat_names in grouped_features.items():
                yield name, inchi, json.dumps([features[name] for name in feat_names])

    def _insert_rows(self, name, rows):
        '''
        Insert rows into feature set tables.

        Args:
            name:
                The feature set name.

            rows:
                An iterator of 2-tuples of InChi strings and JSON strings.
        '''
        LOGGER.info('Inserting %s row(s) into table %s.', len(rows), name)

        with self.conn as conn:
            conn.executemany(f'REPLACE INTO "{name}" VALUES(?, ?)', rows)

    def insert_features(self, items):
        '''
        Insert features into the database.

        Args:
            items:
                An iterator over 3-tuples of (<name>, <inchi>, <features>) where
                <name> is the feature set name, <inchi> is the molecule's InChi,
                and <features> is a dict mapping feature names to their values.

                This must be equivalent to the iterator returned by
                FeatureManager.calculate_features.
        '''
        # Chunk size for database insertion using executemany().
        chunksize = 1000

        all_rows = {}
        for name, inchi, feats in self._parse_inserts(items):
            rows = all_rows.setdefault(name, [])
            rows.append((inchi, feats))
            if len(rows) >= chunksize:
                self._insert_rows(name, rows)
                rows.clear()

        for name, rows in all_rows.items():
            if rows:
                self._insert_rows(name, rows)

    def is_inchi_in_table(self, inchi, name):
        '''
        Check if an InChi is already in a table.

        Args:
            inchi:
                The InChi.

            name:
                The table name.

        Returns:
            A boolean indicating if the InChi exists in the given table.
        '''
        with self.conn as conn:
            return bool(
                conn.execute(
                    f'SELECT 1 FROM "{name}" WHERE {self.INCHI_COLUMN_NAME}=?',
                    [inchi]
                )
            )

    def inchis_in_table(self, name):
        '''
        Return all the InChis in a given table.

        Args:
            name:
                The table name.

        Returns:
            A generator over the InChis in the given table.
        '''
        try:
            with self.conn as conn:
                for row in conn.execute(f'SELECT "{self.INCHI_COLUMN_NAME}" FROM "{name}"'):
                    yield row[0]
        except sqlite3.OperationalError as err:
            if not err.args[0].startswith(self.NO_SUCH_TABLE_ERROR):
                LOGGER.error('%s', err)

    def _execute_to_dicts(self, *args, **kwargs):
        '''
        Execute a query and return a generator that converts each row to a dict.

        Args:
            *args, **kwargs:
                Positional and keyword arguments passed through to execute().

        Returns:
            A generator of dicts.
        '''
        with self.conn as conn:
            conn.row_factory = dict_factory
            yield from conn.execute(*args, **kwargs)

    def get_features(self, inchi, name):
        '''
        Get the features for the InChi from the given table.

        Args:
            inchi:
                The InChi.

            name:
                The table name.

        Returns:
            A dict with the features, or None if the InChi was not found.
        '''
        keys = self.get_feature_names(name)
        if not keys:
            return None

        with self.conn as conn:
            for row in conn.execute(
                f'SELECT {self.FEATURES_COLUMN_NAME} '
                f'FROM "{name}" WHERE {self.INCHI_COLUMN_NAME}=?',
                [inchi]
            ):
                values = json.loads(row[0])
                return dict(zip(keys, values))
        return None

    def _get_join_sql_query(self, names):
        '''
        Get the SQL query to (inner) join the given table names. All tables will
        be joined on the InChi primary key.

        Args:
            names:
                The names of the tables to include.

        Returns:
            The SQL query as a string.
        '''
        first_name = names[0]

        tmp_col_names = ((name, self.FEATURES_COLUMN_NAME) for name in names)
        tmp_col_names = (
            f'"{tname}"."{cname}" AS "{tname}_{cname}"'
            for (tname, cname) in tmp_col_names
        )
        col_names = ', '.join((
            f'"{first_name}"."{self.INCHI_COLUMN_NAME}" AS "{self.INCHI_COLUMN_NAME}"',
            *tmp_col_names
        ))

        sql = f'SELECT {col_names} FROM "{first_name}"'
        for name in names[1:]:
            sql += (
                f' INNER JOIN "{name}" '
                f'ON "{name}"."{self.INCHI_COLUMN_NAME}" = '
                f'"{first_name}"."{self.INCHI_COLUMN_NAME}"'
            )
        LOGGER.debug('SQL INNER JOIN QUERY: %s', sql)
        return sql

    def _get_join_dicts(self, names):
        '''
        Get a generator over dicts of features created by joining the data from
        the given tables.

        Args:
            names:
                The tables to join.

        Returns:
            A generator over the dictionaries.
        '''
        sql = self._get_join_sql_query(names)
        inchi_col_name = self.INCHI_COLUMN_NAME
        suffix = f'_{self.FEATURES_COLUMN_NAME}'
        suffix_len = len(suffix)
        feat_names = self.get_all_feature_names()
        for row in self._execute_to_dicts(sql):
            features = {}
            for key, value in row.items():
                if key.endswith(suffix):
                    name = key[:-suffix_len]
                    features.update(zip(feat_names[name], json.loads(value)))
                elif key == inchi_col_name:
                    features[inchi_col_name] = value
            yield features

    def _filter_inchis(self, rows, inchis=None):
        '''
        Filter rows without selected inchis.j

        Args:
            rows:
                An iterator over dicts as returned by _get_join_dicts().

            inchis:
                An iterable of InChis. Only rows with InChis in this set will be
                included in the output. If None then all rows will be returned.

        Returns:
            A generator over the filtered dicts.
        '''
        if inchis is None:
            yield from rows
            return

        inchis = set(inchis)
        if not inchis:
            return

        for row in rows:
            if row[self.INCHI_COLUMN_NAME] in inchis:
                yield row

    def get_dataframe(self, names, inchis=None):
        '''
        Get a Pandas Dataframe with the features of the given tables.

        Args:
            names:
                The names of the tables to include.

            inchis:
                An optional iterable of InChis.

        Returns:
            The Pandas Dataframe.
        '''
        rows = self._get_join_dicts(names)
        rows = self._filter_inchis(rows, inchis)
        return pd.DataFrame(rows)

    def save_csv(self, path, names, inchis=None):
        '''
        Save features from the given tables to a CSV file.

        Args:
            path:
                The output path.

            names:
                The names of the tables to include.

            inchis:
                An optional iterable of InChis.
        '''
        if isinstance(inchis, pd.Series):
            inchis = inchis.values

        rows = self._get_join_dicts(names)
        rows = self._filter_inchis(rows, inchis)

        LOGGER.info('Saving features to %s', path)
        csv_wrapper = CSVWrapper(path)
        csv_wrapper.write_rows(rows)



#!/usr/bin/env python3
'''
Functions for generating Markdown files.
'''

_SPECIAL_CHARS = r'\`*_{}[]<>()+-.!|'
_TRANS_TABLE = str.maketrans({char: f'\\{char}' for char in _SPECIAL_CHARS})


def escape(text):
    '''
    Escape special symbols in Markdown text.
    '''
    return text.translate(_TRANS_TABLE)



#!/usr/bin/env python3

'''
RDK fingerprint feature calculator.
'''


from rdkit import Chem


from chemfeat.features.calculator import FeatureCalculator


class RDKFPFeatureCalculator(FeatureCalculator):
    '''
    RDK fingerprint feature calculator.

    RDK topological fingerprint calculated with the [RDKit cheminformatics
    library](https://www.rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.RDKFingerprint

    Each feature is a single bit of the feature vector.
    '''
    FEATURE_SET_NAME = 'rdkfp'

    def __init__(self, size: int = 2048):
        '''
        Args:
            size:
                The fingerprint size. It should be 1024, 2048 or 4096.
        '''
        size = int(size)
        valid_sizes = (1024, 2048, 4096)
        if size not in valid_sizes:
            raise ValueError(
                f'Invalid size for fingerprint: {size}. '
                f'Valid sizes: {valid_sizes}'
            )
        self.size = 2048

    @property
    def parameters(self):
        return {'size': self.size}

    def is_numeric(self, _name):
        return False

    def add_features(self, features, molecule):
        fingerprint = Chem.RDKFingerprint(molecule, fpSize=self.size)
        return self.add_fingerprint_features(features, fingerprint)


RDKFPFeatureCalculator.register()



#!/usr/bin/env python3

'''
Functions for working with CSV files.
'''

import csv
import logging
import pathlib


LOGGER = logging.getLogger(__name__)


class CSVWrapper():
    '''
    Wrapper around a CSV file to faciliate appending rows in real-time via an
    iterable of dictionaries.
    '''
    ENCODING = 'utf-8'

    def __init__(self, path, **kwargs):
        self.path = pathlib.Path(path).resolve()
        self.kwargs = kwargs

    @property
    def headers(self):
        '''
        The current headers from the CSV file, or None if they do not exist.

        This assumes that the first row contains the headers.
        '''
        try:
            with self.path.open('r', encoding=self.ENCODING) as csvfile:
                reader = csv.reader(csvfile, **self.kwargs)
                for row in reader:
                    return row
        except FileNotFoundError:
            pass
        return None

    def __iter__(self):
        try:
            with self.path.open('r', encoding=self.ENCODING) as csvfile:
                reader = csv.DictReader(csvfile)
                for row in reader:
                    yield row
        except FileNotFoundError:
            pass

    def write_rows(self, rows, headers=None, append=False):
        '''
        Append rows to the CSV file.

        Args:
            rows:
                A generator of dicts.
        '''
        if append:
            existing_headers = self.headers
            if existing_headers is not None:
                if headers is not None and existing_headers != headers:
                    LOGGER.warning('Re-using existing headers: %s', existing_headers)
                headers = existing_headers

        first_row = None
        try:
            if headers is None:
                first_row = next(rows)
                headers = sorted(first_row)
        except StopIteration:
            LOGGER.warning('Attempting to append 0 rows to %s', self.path)

        mode = 'a' if append else 'w'
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open(mode, encoding='utf-8') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=headers, **self.kwargs)

            if first_row:
                writer.writeheader()
                writer.writerow(first_row)
                csvfile.flush()

            for row in rows:
                writer.writerow(row)
                csvfile.flush()



#!/usr/bin/env python3

'''
SEC fingerprint feature calculator.
'''


from rdkit.Chem import rdMHFPFingerprint


from chemfeat.features.calculator import FeatureCalculator


class SECFPFeatureCalculator(FeatureCalculator):
    '''
    SECFP feature calculator.


    SMILES Extended Connectivity Fingerprint f

    MinHash Fingerprints (MHFP) / SMILES Extended Connectivity Fingerprints
    (SECFP) calculated with [RDKit cheminformatics
    library](https://rdkit.org/docs/source/rdkit.Chem.rdMHFPFingerprint.html).

    > Probst, Daniel, and Jean-Louis Reymond. “A Probabilistic Molecular
    > Fingerprint for Big Data Settings.” Journal of Cheminformatics 10, no. 1
    > (December 18, 2018): 66. https://doi.org/10.1186/s13321-018-0321-8.

    Each feature is a single bit of the feature vector.
    '''
    FEATURE_SET_NAME = 'secfp'

    def __init__(self, size: int = 2048):
        size = int(size)
        valid_sizes = (1024, 2048, 4096)
        if size not in valid_sizes:
            raise ValueError(
                f'Invalid size for fingerprint: {size}. '
                f'Valid sizes: {valid_sizes}'
            )
        self.size = 2048

    @property
    def parameters(self):
        return {'size': self.size}

    def is_numeric(self, _name):
        return False

    def add_features(self, features, molecule):
        fingerprint = rdMHFPFingerprint.MHFPEncoder().EncodeSECFPMol(molecule, length=self.size)
        return self.add_fingerprint_features(features, fingerprint)


SECFPFeatureCalculator.register()



#!/usr/bin/env python3

'''
QED feature calculator.
'''


from rdkit.Chem import QED


from chemfeat.features.calculator import FeatureCalculator


class QEDFeatureCalculator(FeatureCalculator):
    '''
    QED feature calculator.

    Quantitative estimation of drug-likeness features calculated with the [RDKit
    cheminformatics
    library](https://www.rdkit.org/docs/source/rdkit.Chem.QED.html)

    > Bickerton, G. Richard, Gaia V. Paolini, Jérémy Besnard, Sorel Muresan, and
    > Andrew L. Hopkins. “Quantifying the Chemical Beauty of Drugs.” Nature
    > Chemistry 4, no. 2 (February 2012): 90–98.
    > https://doi.org/10.1038/nchem.1243.

    The following features are calculated:

    * ALERTS: The number of structural alerts.
    * ALOGP: The octanol-water partition coefficient.
    * AROM: The number of aromatic rings.
    * HBA: The number of hydrogen bond acceptors.
    * HBD: The number of hydrogen bond donors.
    * MW: The molecular weight.
    * PSA: Polar surface area.
    * ROTB: The number of rotatable bonds.
    '''

    FEATURE_SET_NAME = 'qed'

    def is_numeric(self, _name):
        return True

    def add_features(self, features, molecule):
        properties = QED.properties(molecule)
        features.update(
            (self.add_prefix(name), value)
            for name, value in properties._asdict().items()
        )
        return features


QEDFeatureCalculator.register()



#!/usr/bin/env python3

'''
Manage feature calculations.
'''

import contextlib
import hashlib
import logging
import multiprocessing
import pathlib

from rdkit import Chem
from simple_file_lock import FileLock
import pandas as pd


from chemfeat.database import FeatureDatabase
from chemfeat.features.calculator import FEATURE_CALCULATORS, PREFIX_SEPARATOR
from chemfeat.modules import import_modules


LOGGER = logging.getLogger(__name__)
NAME_KEY = 'name'
INCHI_COLUMN = FeatureDatabase.INCHI_COLUMN_NAME


def import_calculators(paths=None):
    '''
    Import the feature calculator subclasses.

    Args:
        paths:
            An optional list of paths to external modules with feature
            calculator subclasses. The register method of each FeatureCalculator
            subclass should be called by the module after the subclass is
            defined to register the feature calculator globally.

            File paths will import single modules while directory paths will
            recursively import all modules contained in the directory.

    Returns:
        A dict mapping feature set names to feature calculator subclasses.
    '''
    search_paths = [
        pathlib.Path(__file__).resolve().parent / 'calculators'
    ]
    if paths:
        search_paths.extend(paths)

    import_modules(
        search_paths,
        path_log_msg='Feature calculator search path: %s'
    )
    return FEATURE_CALCULATORS.copy()


def _calculate_features(inchi_and_feat_calcs):
    '''
    Internal function for calculating features with a multiprocessing pool.

    Args:
        inchi_and_feat_calcs:
            A 2-tuple consisting of an InChi string and a list of feature
            calculators.

    Returns:
        The InChi and the dict of features.
    '''
    inchi, feat_calcs = inchi_and_feat_calcs
    features = {}
    LOGGER.debug('Converting %s to molecule', inchi)
    molecule = Chem.inchi.MolFromInchi(inchi)
    if molecule is None:
        LOGGER.error('Failed to convert InChi to molecule object: %s', inchi)
        return None, features
    for calc in feat_calcs:
        # TODO
        # If there is a problem using the same FeatureCalculator objects with
        # multiprocessing, instantiate new objects using the class and
        # parameters.
        calc.add_features(features, molecule)
    return inchi, features


class FeatureManager():
    '''
    Calculate features from InChis and save the results to a database.
    '''
    def __init__(self, feature_database, features):
        '''
        Args:
            feature_database:
                A FeatureDatabase object.

            features:
                An iterable of dicts. Each dict must contain the key "name"
                which designates the feature set to use. All other key-value
                pairs in the dict will be interpretted as parameters for the
                designated feature set.

        Raises:
            See parse().
        '''
        self.feature_database = feature_database
        self.parse(features)
        self.inchis = None
        self.molecules = None

    def parse(self, features):
        '''
        Parse feature specifications.

        Args:
            features:
                Same as __init__().

        Raises:
            ValueError:
                A required key was absent or invalid.
        '''
        all_feat_calcs = import_calculators()
        feat_calcs = {}
        for parameters in features:
            try:
                name = parameters.pop(NAME_KEY)
            except KeyError as err:
                raise ValueError('Feature specification lacks "name" key.') from err
            try:
                calc_cls = all_feat_calcs[name]
            except KeyError as err:
                raise ValueError(f'Unrecognized feature set: {name}') from err
            calc = calc_cls(**parameters)
            feat_calcs[calc.identifier] = calc
        self.feature_calculators = feat_calcs

    def get_feature_parameters(self):
        '''
        Get the feature parameters from the currently configured features.

        Returns:
            A list of features as accepted by __init__().

        '''
        for calc in self.feature_calculators.values():
            params = calc.parameters
            params[NAME_KEY] = calc.FEATURE_SET_NAME
            yield params

    def is_numeric(self, feature_name):
        '''
        Check if a feature is numeric as opposed to categorical.

        Args:
            feature_name:
                The name of the feature.

        Returns:
            True if the feature is numeric, False if it is categorical.
        '''
        set_name, name = feature_name.split(PREFIX_SEPARATOR, 1)
        return self.feature_calculators[set_name].is_numeric(name)

    def numeric_mask(self, feature_names):
        '''
        Get a mask for the numeric features.

        Args:
            feature_names:
                An iterable of feature names.

        Returns:
            A Pandas series of booleans that serve as a mask.
        '''
        if not isinstance(feature_names, pd.Series):
            feature_names = pd.Series(feature_names)
        return feature_names.apply(self.is_numeric)

    def categoric_mask(self, feature_names):
        '''
        Get a mask for categoric features. This is just a wrapper around
        numeric_mask() that accepts the same arguments and inverts the mask.
        '''
        return ~self.numeric_mask(feature_names)

    @property
    def feature_set_string(self):
        '''
        A unique string representing the feature set.
        '''
        long_identifier = ' '.join(calc.identifier for calc in self.feature_calculators.values())
        return hashlib.sha256(long_identifier.encode('utf-8')).hexdigest()

    def filter_feature_specs(self, inchis):
        '''
        Filter feature specifications based on what is already in the database.
        This assumes that the existing database tables contain the expected
        data, which may not be the case if the feature sets have changed.

        Args:
            inchis:
                An iterable of target InChis.

        Returns:
            A filtered list of 2-tuples mapping InChis to the missing feature
            specifications.
        '''
        precalculated = []
        for calc in self.feature_calculators.values():
            existing_inchis = set(self.feature_database.inchis_in_table(calc.identifier))
            precalculated.append((existing_inchis, calc))

        for inchi in inchis:
            feat_calcs = []
            for existing_inchis, calc in precalculated:
                if inchi not in existing_inchis:
                    feat_calcs.append(calc)
            # Only yield the InChi if there are uncalculated features.
            if feat_calcs:
                yield inchi, feat_calcs

    def calculate_features(
        self,
        inchis,
        output_path=None,
        return_dataframe=False,
        n_jobs=-1
    ):
        '''
        Get the path to a CSV file with the current feature set. If the file
        does not exist, it will be created.

        Args:
            inchis:
                An iterable of InChi strings.

            output_path:
                An optional output path for saving the results to a CSV file.

            return_dataframe:
                If True, return a Pandas dataframe with the results.

            n_jobs:
                The number of jobs to use when calculating features.
        '''
        if output_path:
            output_path = pathlib.Path(output_path).resolve()
            output_ctxt = FileLock(output_path)
        else:
            output_ctxt = contextlib.nullcontext(output_path)

        with FileLock(self.feature_database.path), output_ctxt:
            if n_jobs < 1:
                n_jobs = multiprocessing.cpu_count()

            with multiprocessing.Pool(n_jobs) as pool:
                # Convert to a list to avoid passing generator with SQlite
                # database reference to threads/processes, which raises an
                # exception.
                features = pool.imap_unordered(
                    _calculate_features,
                    list(self.filter_feature_specs(inchis))
                )

                features = (
                    (inchi, feats)
                    for inchi, feats in features
                    if inchi is not None and feats
                )
                self.feature_database.insert_features(features)

            feature_set_names = [calc.identifier for calc in self.feature_calculators.values()]
            if output_path:
                self.feature_database.save_csv(output_path, feature_set_names, inchis=inchis)
            if return_dataframe:
                return self.feature_database.get_dataframe(feature_set_names, inchis=inchis)
        return None



#!/usr/bin/env python3

'''
RDK descriptor feature calculator.
'''


from rdkit.Chem import Descriptors


from chemfeat.features.calculator import FeatureCalculator


class RDKDescFeatureCalculator(FeatureCalculator):
    '''
    RDK descriptor feature calculator.

    Various chemical descriptors calculated with the RDKit cheminformatics
    library: https://www.rdkit.org/docs/source/rdkit.Chem.Descriptors.html

    The following features are calculated:

    * FpDensityMorgan1
    * FpDensityMorgan2
    * FpDensityMorgan3
    * MaxAbsPartialCharge
    * MaxPartialCharge
    * MinAbsPartialCharge
    * MinPartialCharge
    * NumRadicalElectrons
    * NumValenceElectron
    '''
    FEATURE_SET_NAME = 'rdkdesc'

    def is_numeric(self, _name):
        return True

    def add_features(self, features, molecule):
        features.update(
            (self.add_prefix(name), getattr(Descriptors, name)(molecule))
            for name in (
                'FpDensityMorgan1',
                'FpDensityMorgan2',
                'FpDensityMorgan3',
                'MaxAbsPartialCharge',
                'MaxPartialCharge',
                'MinAbsPartialCharge',
                'MinPartialCharge',
                'NumRadicalElectrons',
                'NumValenceElectrons'
            )
        )
        return features


RDKDescFeatureCalculator.register()
